module Screen(
    Packing(..),
    BasicWidget(..),
    Attribute(..),
    Changeable(..),
    Readable(..),
    doRender,

    RenderContext,
    touchRenderContext,
    newRenderContext,
    setRenderContext,
    tryDrawRenderContext,
    resizeRenderContext,
    setRenderWidget,
    keyRenderContext,
    Cache,
    newCache,
    cacheInvalidate,
    cacheSignal,
    readCache,
    newFakeWidget,
    newAlternate,
    newSVar,
    writeSVar,
    modifySVar,
    readSVar,
    newSVarWidget,
    Widget, SVar,
    Signal(..),
    svarSignal

    ) where

import Control.Exception as E
import Control.Monad.Trans
import Data.Bits
import Data.Unique
import Control.Concurrent
import Monad

import ErrorLog
import Curses
import Format
import GenUtil



data Canvas = Canvas {window :: !Window, origin :: (Int,Int), widgetOrigin :: (Int,Int), bounds :: (Int,Int)  }

data Widget = Widget {
    render :: Canvas -> IO (), 
    getBounds :: IO (Int,Int), 
    changedSignal :: Signal (),
    processKey :: Key -> IO Bool
    }

data Alternate = Alternate !(MVar (Widget,IO ())) !(Signal ())

data Packing = NoExpand | Expand | ExpandFill
data Attribute = AttrBold | AttrBlink | AttrDim | AttrReverse | AttrUnderline | AttrColor String String

instance Show Canvas where
    show c = fmtSs "<canvas:origin=%s,bounds=%s>" [show (origin c), show (bounds c)]


class BasicWidget w where
    widgetEmpty :: w
    widgetEmptySized :: Float -> Float -> w
    widgetCenter :: w -> w
    widgetSimpleFrame :: w -> w
    widgetVerticalBox :: Bool -> [(Packing, w)] -> w
    widgetHorizontalBox :: Bool -> [(Packing, w)] -> w
    widgetText :: String -> w
    widgetAttr :: [Attribute] -> w -> w
    widgetScroll :: w -> IO w
    -- project :: IO w -> w


instance BasicWidget Widget where
    widgetEmpty = emptyWidget
    widgetEmptySized x y = emptySizedWidget (floor (x + 0.5)) (floor (y + 0.5))
    widgetCenter w = centerWidget w 
    widgetSimpleFrame w = boxWidget w
    widgetVerticalBox True ws = homVerticalBox ws
    widgetVerticalBox False ws = verticalBox ws []
    widgetHorizontalBox True ws = homHorizontalBox ws
    widgetHorizontalBox False ws = horizontalBox ws []
    widgetText s = staticText s
    widgetAttr al w = attrWidget (foldl (flip ($)) attr0 (map ma al)) w where
	ma AttrBlink a = a `setBlink` True
	ma AttrBold a = a  `setBold` True
	ma AttrDim a = a  `setDim` True
	ma AttrReverse a = a  `setReverse` True
	ma AttrUnderline a = a `setUnderline` True
	ma _ a = a
    widgetScroll = newScrollWidget


instance BasicWidget () where
    widgetEmpty = ()
    widgetEmptySized x y = ()
    widgetCenter w = ()
    widgetSimpleFrame w = ()
    widgetVerticalBox _ ws = ()
    widgetHorizontalBox _ ws = ()
    widgetText s = ()
    widgetAttr _ _ = ()
    widgetScroll _ = return ()



isExpand NoExpand = False
isExpand _ = True

isFill ExpandFill = True
isFill _ = False


data RenderContext = RenderContext { 
    needsRedraw :: MVar Bool, 
    needsResize :: MVar Bool, 
    drawingStuff :: MVar (IO (), Key -> IO Bool, IO ())
    }

newRenderContext = do
    x <- newMVar False
    y <- newMVar False
    z <- newMVar (return (), \_ -> return False, return ())
    return RenderContext {needsRedraw = x, needsResize = y, drawingStuff = z }

touchRenderContext rc = swapMVar (needsRedraw rc) True >> return ()

resizeRenderContext rc action = swapMVar (needsResize rc) True >>= \x -> unless x action >> return ()

tryDrawRenderContext rc = do
    x <- swapMVar (needsRedraw rc) False
    y <- swapMVar (needsResize rc) False
    withMVar (drawingStuff rc) $ \(z,_,_) -> when (x || y) $ do
	erase
	when y (attempt endWin >> refresh)
	z
	refresh

setRenderContext rc dr = do
    modifyMVar_ (drawingStuff rc)  $ \(_,_,final) -> do
    final
    touchRenderContext rc 
    return (dr,\_ -> return False, return ())

setRenderWidget rc w = do
    modifyMVar_ (drawingStuff rc)  $ \(_,_,final) -> do
    final
    nf <- connect (changedSignal w) (\_ -> touchRenderContext rc)
    touchRenderContext rc  
    return (wdr, processKey w, nf) where
	wdr = do
	    (ys,xs) <- scrSize
	    let c = (Canvas {window = stdScr, origin = (0,0), widgetOrigin = (0,0), bounds = (xs,ys)})  
	    eannM ("doRender " ++ show c) $ render w c

keyRenderContext :: RenderContext -> Key -> IO Bool
keyRenderContext rc KeyResize = return True
keyRenderContext rc k = do 
    (_,pks,_) <- readMVar (drawingStuff rc)
    pks k

doRender w = do
    (ys,xs) <- scrSize
    erase 
    let c = (Canvas {window = stdScr, origin = (0,0), widgetOrigin = (0,0), bounds = (xs,ys)})  
    eannM ("doRender " ++ show c) $ render w c
    refresh

    



-------------------
-- Drawing Routines
-------------------

canvasTranslate :: Canvas -> (Int,Int) -> Canvas
canvasTranslate canvas@(Canvas {origin = (x,y), bounds = (xb,yb)}) (xt,yt) = assert (xt > 0 && yt > 0) $ canvas {origin = (x + xt, y + yt), bounds = (max (xb - xt) 0, max (yb - yt) 0)}


canvasWTranslate :: Canvas -> (Int,Int) -> Canvas
canvasWTranslate canvas@(Canvas {widgetOrigin = (x,y)}) (xt,yt) = canvas {widgetOrigin = (x + xt, y + yt)}

withAttr :: Canvas -> Attr -> IO a -> IO a
withAttr canvas attr action = E.bracket (wAttrGet w) (wAttrSet w) f where
    f (a,p) = wAttrSet w (a .|. attr,p) >> action
    w = window canvas


drawString :: Canvas -> String -> IO ()
drawString canvas s = eannM (fmtSs "drawString %s %s" [(show canvas), s]) $ ds ml yb (origin canvas) where
	ds (s:ls) yb (x,y) | yb > 0 = attempt (mvWAddStr (window canvas) y x (take xb s)) >> ds ls (yb - 1) (x,y + 1)
	ds _ _ _ = return ()
	(xb,yb) = bounds canvas
	ls = lines s
	(wx,wy) = widgetOrigin canvas
	ml = map (mv ' ' wx) (mv "" wy ls)
	mv _ 0 ss = ss
	mv _ n ss | n > 0 = drop n ss
	mv ec n ss = replicate (0 - n) ec ++ ss


boundCanvas :: Maybe Int -> Maybe Int -> Canvas -> Canvas 
boundCanvas mx my canvas = (canvas {bounds = (nx,ny)}) where
    (cx,cy) = bounds canvas
    nx = maybe cx (min cx) mx
    ny = maybe cy (min cy) my


-- rendering routines

renderCentered :: Widget -> Canvas -> IO ()
renderCentered w canvas = do
	(xs,ys) <- getBounds w
	let (xs',ys') =  bounds canvas
	    f cs ws | ws >= cs = 0
	    f cs ws = (cs - ws) `div` 2
	    nc = (canvasTranslate canvas (f xs' xs,f ys' ys))
	renderChild w nc {bounds=(min xs $ fst $ bounds nc ,min ys $ snd $ bounds nc)} 

renderBounded :: Widget -> Canvas -> Maybe Int -> Maybe Int -> IO ()
renderBounded w canvas mx my = renderChild w (canvas {bounds = (nx,ny)}) where
    (cx,cy) = bounds canvas
    nx = maybe cx (min cx) mx
    ny = maybe cy (min cy) my

renderChild :: Widget -> Canvas -> IO ()
renderChild _ (Canvas {bounds = (xb,yb)}) | xb == 0 || yb == 0 = return ()
renderChild w c = render w c

----------
-- Widgets
----------



emptyWidget = Widget {
    render = const (return ()), 
    getBounds = return (0,0), 
    changedSignal = nonSignal,
    processKey = \_ -> return False
    }
emptySizedWidget x y =  emptyWidget {getBounds = return (x,y)}


childrenWidget ws = emptyWidget {changedSignal = signalCondenser (map changedSignal ws), processKey = pk} where
    pk k = tk ws k
    tk [] _ = return False
    tk (w:ws) k = do
	b <- processKey w k
	if b then return True else tk ws k
-- layout widgets

-- | centers child widget, causes child to not use any extra space.
centerWidget :: Widget -> Widget
centerWidget w = w {render = rst} where
    rst canvas = do 
	(xs,ys) <- getBounds w
	let (xs',ys') =  bounds canvas
	renderChild w (canvasTranslate canvas (f xs' xs,f ys' ys)) {bounds=(min xs $ fst $ bounds canvas ,min ys $ snd $ bounds canvas)} where
	    f cs ws | ws >= cs = 0
	    f cs ws = (cs - ws) `div` 2


-- General Widgets


newScrollWidget :: Widget -> IO Widget
newScrollWidget w = do
    yr <- newMVar 0
    sig <- newSignal
    connect (changedSignal w) (signal sig)
    let rst canvas = do
	    y <- readMVar yr
	    renderChild w (canvasWTranslate canvas (0,y))  where
	pk (KeyChar 'j') = modifyMVar_ yr (\x -> return $ x + 1) >> keyDone
	pk KeyDown = modifyMVar_ yr (\x -> return $ x + 1) >> keyDone
	pk (KeyChar 'k') = modifyMVar_ yr (\x -> return $ x - 1) >> keyDone
	pk KeyUp = modifyMVar_ yr (\x -> return $ x - 1) >> keyDone
	pk KeyHome = swapMVar yr 0 >> keyDone
	pk (KeyChar 'g') = swapMVar yr 0 >> keyDone
	pk k = processKey w k
	keyDone = modifyMVar_ yr (\x -> return  (max 0 x)) >> signal sig () >> return True

    return w {processKey = pk, render = rst, changedSignal = sig}
    

-- | draw box around child widget
boxWidget :: Widget -> Widget
boxWidget w = w {render = rst, getBounds = gb} where
    gb = getBounds w >>= \(x,y) -> return (x + 2, y + 2)
    rst canvas@(Canvas {bounds = (xb,yb)}) = do
	let tb = '+' : (replicate (xb - 2) '-' ++ "+")
	    bc c@(Canvas {bounds = (xb,yb)}) = c {bounds = (max (xb - 1) 0, max (yb - 1) 0)}
	renderChild w (canvasTranslate (bc canvas) (1,1))	
	drawString canvas tb
	drawString (canvasTranslate canvas (0,yb - 1)) tb
	sequence_ [drawString (canvasTranslate canvas (xt,yt)) "|"| yt <- [1..yb - 2], xt <- [0,xb - 1]]


{-
widgetMinBBox :: Maybe Int -> Maybe Int -> Widget -> Widget
widgetMaxBBox :: Maybe Int -> Maybe Int -> Widget -> Widget
widgetExactBBox :: Maybe Int -> Maybe Int -> Widget -> Widget
-}
    
staticText s = emptyWidget {render = rst, getBounds = return bounds } where
    bounds = (maximum (0:map length ls),length ls)
    ls = lines s
    rst canvas = drawString canvas s 


filledSizedWidget :: Int -> Int -> Char -> Widget
filledSizedWidget 0 y _ = (emptySizedWidget 0 y) 
filledSizedWidget x 0 _ = (emptySizedWidget x 0) 
filledSizedWidget x y c = (emptySizedWidget x y) {render = rnd} where
    rnd canvas = eannM (fmt "filledSizedWidget %i %i %/" [fi x, fi y, fa c]) $ ds canvas (min yb y)
     where
	ds _ 0 = return ()
	ds canvas yb = do 
	    drawString canvas (replicate (min xb x) c) 
	    ds (canvasTranslate canvas (0,1)) (yb - 1) 
	(xb,yb) = bounds canvas
	


verticalBox :: [(Packing,Widget)] -> [(Packing,Widget)] -> Widget
verticalBox starts ends = (childrenWidget aw) {render = rd, getBounds = gb} where
    awpb = starts ++ ends
    awp = if any isExpand $ fsts awpb then awpb else starts ++ [(Expand,emptyWidget)] ++ ends
    aw = snds awp
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (maximum xs, sum ys)
    rd canvas = do
	let (xb,yb) = bounds canvas
	(xs,ys) <- gb
	let ne = length (filter isExpand $ fsts awp)
	    el = if yb > ys then splitSpace (yb - ys) ne else replicate ne 0
	    f ((e:es),canvas) (ps,w) | isExpand ps = do
		(wbx,wby) <- getBounds w
		if isFill ps then 
		    renderChild w (boundCanvas Nothing (Just (wby + e)) canvas)
			else renderCentered w (boundCanvas Nothing (Just (wby + e)) canvas)
		return (es,canvasTranslate canvas (0,wby + e))
	    f (es,canvas) (ps,w) = do
		(wbx,wby) <- getBounds w
		--renderCentered w (boundCanvas Nothing (Just wby) canvas)
		renderChild w (boundCanvas Nothing (Just wby) canvas)
		return (es,canvasTranslate canvas (0,wby))
	foldlM_ f (el,canvas) awp


horizontalBox :: [(Packing,Widget)] -> [(Packing,Widget)] -> Widget
horizontalBox starts ends = (childrenWidget aw) {render = rd, getBounds = gb } where
    awpb = starts ++ ends
    awp = if any isExpand $ fsts awpb then awpb else starts ++ [(Expand,emptyWidget)] ++ ends
    aw = snds awp
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (sum xs, maximum ys)
    rd canvas = do
	let (xb,yb) = bounds canvas
	(xs,ys) <- gb
	let ne = length (filter isExpand $ fsts awp)
	    el = if xb > xs then splitSpace (xb - xs) ne else replicate ne 0
	    f ((e:es),canvas) (ps,w) | isExpand ps = do
		(wbx,wby) <- getBounds w
		if isFill ps then 
		    renderChild w (boundCanvas (Just (wbx + e)) Nothing canvas)
			else renderCentered w (boundCanvas (Just (wbx + e)) Nothing canvas)
		return (es,canvasTranslate canvas (wbx + e,0))
	    f (es,canvas) (ps,w) = do
		(wbx,wby) <- getBounds w
		--renderCentered w (boundCanvas (Just wbx) Nothing canvas)
		renderChild w (boundCanvas (Just wbx) Nothing canvas)
		return (es,canvasTranslate canvas (wbx,0))
	foldlM_ f (el,canvas) awp

homVerticalBox :: [(Packing,Widget)] -> Widget
homVerticalBox awp = (childrenWidget aw) {render = rd, getBounds = gb } where
    aw = snds awp
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (maximum xs, (maximum ys) * length aw)
    rd canvas = do
	let (xb,yb) = bounds canvas
	    el = splitSpace yb (length aw) 
	    f ([],_) _ = error "no space to split"
	    f ((e:es),canvas) (ps,w)  = do
		if isFill ps then 
		    renderChild w (boundCanvas Nothing (Just e) canvas)
			else renderCentered w (boundCanvas Nothing (Just e) canvas)
		return (es,canvasTranslate canvas (0,e))
	foldlM_ f (el,canvas) awp


homHorizontalBox :: [(Packing,Widget)] -> Widget
homHorizontalBox awp = (childrenWidget aw) {render = rd, getBounds = gb } where
    aw = snds awp
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (maximum xs * length aw, maximum ys)
    rd canvas = do
	let (xb,yb) = bounds canvas
	    el = splitSpace xb (length aw) 
	    f ([],_) _ = error "no space to split"
	    f ((e:es),canvas) (ps,w)  = do
		if isFill ps then 
		    renderChild w (boundCanvas (Just e) Nothing canvas)
			else renderCentered w (boundCanvas (Just e) Nothing canvas)
		return (es,canvasTranslate canvas (e,0))
	foldlM_ f (el,canvas) awp

splitSpace m n = f (m `mod` n) $ replicate n (m `div` n) where
    f 0 xs = xs
    f n (x:xs) = (x + 1) : f (n - 1) xs
    f n' [] = errorf "splitSpace %i %i (f %i []): this can't happen." [fi m, fi n, fi n']


newFakeWidget :: ((Int,Int) -> (Int,Int) -> IO ()) -> IO Widget
newFakeWidget f = do
    sig <- newSignal
    let rnd canvas = f (origin canvas) (bounds canvas)
    return $ emptyWidget { changedSignal = sig, render = rnd}
	
newSVarWidget :: (Readable c, Changeable (c a)) => c a -> (a -> Widget) -> Widget
newSVarWidget sv f = emptyWidget {render = rw, getBounds = gb, changedSignal = signalChanged sv} where
    rw canvas = do
	v <- readVal sv
	render (f v) canvas
    gb = do
	v <- readVal sv
	getBounds (f v)
    

newAlternate :: Widget -> IO (Widget, Alternate)
newAlternate w = do
    sig <- newSignal
    final <- connect (changedSignal w) (\() -> signal sig ())
    r <- newMVar (w,final)
    let w = emptyWidget {render = rw, getBounds = gb, processKey = pk } 
	rw canvas = do
	    (w,_) <- readMVar r
	    renderChild w canvas
	gb = do
	    (w,_) <- readMVar r
	    getBounds w 
	pk k = do
	    (w,_) <- readMVar r
	    processKey w k
    return (w,Alternate r sig)
    

switchAlternate :: Alternate -> Widget -> IO ()
switchAlternate (Alternate mv sig) w = (modifyMVar_ mv $ \(_,final) -> 
	(final >> connect (changedSignal w) (\() -> signal sig ()) >>= \nf -> return (w,nf)))
	    >> signal sig ()


attrWidget :: Attr -> Widget -> Widget
attrWidget a w = w {render = nr} where
    nr canvas = withAttr canvas a (renderChild w  canvas)

--------------------
-- Signals and SVars
--------------------

data Signal a = Signal {
    connect :: MonadIO m =>  (a -> IO ()) -> m (IO ()),
    disconnectAll :: MonadIO m => m (),
    signal :: MonadIO m => a -> m ()
    }


-- signals

signal_ :: Signal a -> Signal ()
signal_ s = s {connect = nc, signal = ns } where 
    nc f = connect s (\_ -> f ())
    ns _ = signal s undefined

newSignal :: MonadIO m => m (Signal a)
newSignal = do
    mv <- liftIO $ newMVar ([]::[(Unique,a -> IO ())])
    let c action = do
	    u <- liftIO newUnique
	    liftIO (modifyMVar_ mv (\as -> return ((u,action):as))) 
	    return (modifyMVar_ mv (\as -> return [ v|v@(u',_) <- as, u' /= u])) 
	d :: MonadIO m => m ()
	d = (liftIO $ swapMVar mv []) >> return ()
	s v =  do
	    cl <- liftIO $ readMVar mv
	    liftIO $ mapM_ (($ v) . snd) cl
    return $ Signal {connect = c, disconnectAll = d, signal = s} 


    
nonSignal = Signal {
    connect = \_ -> return (return ()),
    disconnectAll = return (),
    signal = \_ -> return ()
    }


signalCondenser :: [Signal a] -> (Signal a)
signalCondenser sl = nonSignal {
    connect = \f -> mapM (flip connect f) sl >>= \x -> return (sequence_ x),
    disconnectAll = mapM_ disconnectAll sl
    }


-- Cache
data Cache a = Cache {cacheValue :: !(MVar (Maybe a)), cacheFill :: !(IO a), cacheSignal :: !(Signal ())}

newCache :: IO a -> IO (Cache a)
newCache fill = do
    mv <- newMVar Nothing
    s <- newSignal
    return $ Cache {cacheValue = mv, cacheSignal = s, cacheFill = fill}

readCache :: Cache a -> IO a
readCache (Cache {cacheValue = mv, cacheFill = fill}) = modifyMVar mv $ \x -> case x of 
    v@(Just a) -> return (v,a)
    Nothing -> fill >>= \a -> return (Just a, a)

cacheInvalidate :: Cache a -> IO ()
cacheInvalidate c = swapMVar (cacheValue c) Nothing >> signal (cacheSignal c) () >> return ()

-- SVars


data SVar a = SVar (Signal a) (MVar a)



svarSignal :: SVar a -> Signal a
svarSignal (SVar s _) = s

newSVar :: MonadIO m => a -> m (SVar a)
newSVar v = do
    s <- newSignal 
    mv <- liftIO $ newMVar v
    return $ SVar s mv
writeSVar :: MonadIO m => (SVar a) -> a -> m ()
writeSVar (SVar s mv) v = do
    liftIO $ modifyMVar_ mv (const (return v))
    signal s v
modifySVar :: MonadIO m => (SVar a) -> (a -> a) -> m ()
modifySVar (SVar s mv) f = do
    v <- liftIO $ modifyMVar mv (\x -> (let y = f x in return (y,y)))
    signal s v
readSVar :: MonadIO m => (SVar a) -> m a
readSVar (SVar _ mv) = liftIO $ readMVar mv  

{-
modifySVar_ :: MonadIO m => (SVar a) -> (a -> IO a) -> m ()
modifySVar_ (SVar s mv) f = do
    v <- liftIO $ modifyMVar mv (\x -> (let y = f x in return (y,y)))
    signal s v
-}
    

class Readable c where
    readVal :: MonadIO m => c a -> m a


class Changeable a where
    signalChanged :: a -> Signal ()

instance Changeable Widget where
    signalChanged = changedSignal

instance Changeable (Cache a) where
    signalChanged = cacheSignal

instance Changeable (SVar a) where
    signalChanged = signal_ . svarSignal

instance Readable Cache  where
    readVal x = liftIO $ readCache x

instance Readable SVar where
    readVal x = liftIO $ readSVar x


{-
class Frillable f where
    combine :: f a -> f b -> f (a,b)
    void :: f a -> f ()
    observe :: f a -> a -> IO b -> f b
-}

{- Attic

newtype KeyHandler = KeyHandler (Key -> IO KeyHandler)
unKeyHandler (KeyHandler kh) = kh

{-# NOINLINE loopExit_r #-}
loopExit_r = unsafePerformIO $ newMVar False

exitLoop :: IO ()
exitLoop = swapMVar loopExit_r True >> return ()

runLoop :: KeyHandler -> Widget -> IO ()
runLoop kh w = gProgram $ \canvas ->  rp canvas >> loop canvas kh where
	loop canvas kh = do
	    hc <- hasChanged w
	    when hc $ rp canvas
	    k <- getCh
	    kh <- unKeyHandler kh k
	    c <- readMVar loopExit_r 
	    unless c $ loop canvas kh
	gProgram :: (Canvas -> IO a) -> IO a 
	gProgram action = E.bracket_ initCurses endWin $ do
	   ss <- scrSize 
	   action $ Canvas {window = stdScr, origin = (0,0), bounds = ss}
	rp canvas = erase >> render w canvas >> refresh

------------
-- Test Code
------------

nkh a = KeyHandler f where
    f (KeyChar 'q') = exitLoop >> (return $ nkh a)
    f key@(KeyF i) = do
	switchAlternate a (simpleVerticalBox $ replicate i (staticText (show key)))
	(return $ nkh a)
    f key = do
	switchAlternate a (staticText (show key))
	return $ nkh a
    
	

screenMain = do
    mainThd <- myThreadId
    Posix.installHandler Posix.sigINT (Posix.Catch (killThread mainThd)) Nothing
    (w,a) <- newAlternate
    switchAlternate a (staticText "Sup?")
    runLoop (nkh a) (simpleVerticalBox [w,staticText "Is the keystroke."])

-------------
-- Disposable
-------------


class Disposable a where 
    dispose :: a -> IO ()
    
withDisposable :: (MonadIO m, Disposable a) => m a -> (a -> m b) -> m b
withDisposable c a = do
    v <- c
    r <- a v
    liftIO $ dispose v
    return r
    

-}
